#!/usr/bin/env python
###############################################################################
#
# Purpose:  CityJSON OGR driver
# Author:   Even Rouault <even dot rouault at spatialys.com>
#
###############################################################################
# Copyright (c) 2019, Even Rouault <even dot rouault at spatialys.com>
# SPDX-License-Identifier: MIT
###############################################################################

# Metadata parsed by GDAL C++ code at driver pre-loading, starting with '# gdal: '
# gdal: DRIVER_NAME = "CityJSON"
# gdal: DRIVER_SUPPORTED_API_VERSION = [1]
# gdal: DRIVER_DCAP_VECTOR = "YES"
# gdal: DRIVER_DCAP_VIRTUALIO = "YES"
# gdal: DRIVER_DMD_LONGNAME = "CityJSON"
# gdal: DRIVER_DMD_EXTENSIONS = "json"

import json
import os

from gdal_python_driver import BaseDataset, BaseDriver, BaseLayer


class Layer(BaseLayer):
    def __init__(self, filename, content):
        self.content = content
        self.name = os.path.splitext(os.path.basename(filename))[0]
        self.fields = [
            {"name": "id"},
            {"name": "type"},
        ]

        self.count = 0
        set_attrs = set()
        for key_value in content["CityObjects"].items():
            value = key_value[1]
            if "attributes" in value:
                for kv in value["attributes"].items():
                    if not kv[0] in set_attrs:
                        set_attrs.add(kv[0])
                        v = kv[1]
                        t = "String"
                        if isinstance(v, int):
                            t = "Integer"
                        elif isinstance(v, float):
                            t = "Real"
                        elif "Date" in kv[0]:
                            t = "Date"
                        self.fields.append({"name": kv[0], "type": t})
            for geom in value["geometry"]:
                if geom["type"] in ("Solid", "MultiSurface"):
                    self.count += 1

        md = content["metadata"]
        srs = md["referenceSystem"] if "referenceSystem" in md else None
        if not srs and "crs" in md:
            if isinstance(md["crs"], dict) and "epsg" in md["crs"]:
                srs = "EPSG:" + str(md["crs"]["epsg"])
        self.geometry_fields = [{"type": "MultiPolygonZ", "srs": srs}]
        if "transform" in content:
            self.translate = content["transform"]["translate"]
            self.scale = content["transform"]["scale"]
        else:
            self.translate = [0, 0, 0]
            self.scale = [1, 1, 1]
        self.vertices = content["vertices"]

    def extent(self, force):
        md = self.content["metadata"]
        if "geographicalExtent" in md:
            bbox = md["geographicalExtent"]
        else:
            bbox = md["bbox"]
        return [bbox[0], bbox[1], bbox[3], bbox[4]]

    def feature_count(self, force):
        return self.count

    def __iter__(self):
        fid = 1
        vertices = self.vertices
        scale = self.scale
        translate = self.translate
        for key_value in self.content["CityObjects"].items():
            value = key_value[1]
            for geom in value["geometry"]:
                geom_in = None
                boundaries = geom["boundaries"]
                if geom["type"] == "Solid":
                    assert len(boundaries) == 1, (len(boundaries), key_value)
                    geom_in = boundaries[0]
                elif geom["type"] == "MultiSurface":
                    geom_in = boundaries
                if geom_in:
                    out_mpoly = []
                    for poly in geom_in:
                        out_poly = []
                        for ring in poly:
                            out_ring = []
                            for vertex_idx in ring:
                                v = vertices[vertex_idx]
                                c = [v * s + t for v, s, t in zip(v, scale, translate)]
                                out_ring.append(c)
                            out_ring.append(out_ring[0])
                            out_poly.append(out_ring)
                        out_mpoly.append(out_poly)
                    properties = {"id": key_value[0], "type": value["type"]}
                    if "attributes" in value:
                        properties.update(value["attributes"])

                    wkt = "MULTIPOLYGON Z("
                    first_poly = True
                    for poly in out_mpoly:
                        if not first_poly:
                            wkt += ","
                        else:
                            first_poly = False
                        wkt += "("
                        first_ring = True
                        for ring in poly:
                            if not first_ring:
                                wkt += ","
                            else:
                                first_ring = False
                            wkt += "("
                            first_vertex = True
                            for v in ring:
                                if not first_vertex:
                                    wkt += ","
                                else:
                                    first_vertex = False
                                wkt += "%.18g %.18g %.18g" % (v[0], v[1], v[2])
                            wkt += ")"
                        wkt += ")"
                    wkt += ")"

                    f = {
                        "type": "OGRFeature",
                        "id": fid,
                        "fields": properties,
                        "geometry_fields": {"": wkt},
                    }
                    yield f
                    fid += 1


class Dataset(BaseDataset):
    def __init__(self, filename, content):
        self.layers = [Layer(filename, content)]


class Driver(BaseDriver):
    def identify(self, filename, first_bytes, open_flags, open_options={}):
        return (
            b'"type":"CityJSON"' in first_bytes
            or b'"type": "CityJSON"' in first_bytes
            or first_bytes.startswith(b'{"CityObjects":{')
        )

    def open(self, filename, first_bytes, open_flags, open_options={}):
        if not self.identify(filename, first_bytes, open_flags):
            return None
        if filename.startswith("/vsi"):
            from osgeo import gdal

            f = gdal.VSIFOpenL(filename, "rb")
            if not f:
                return None
            gdal.VSIFSeekL(f, 0, 2)
            length = gdal.VSIFTellL(f)
            gdal.VSIFSeekL(f, 0, 0)
            try:
                content = json.loads(gdal.VSIFReadL(1, length, f))
            finally:
                gdal.VSIFCloseL(f)
        else:
            with open(filename, "rt") as f:
                content = json.loads(f.read())
        return Dataset(filename, content)



#!/usr/bin/env python
# SPDX-License-Identifier: MIT
# Copyright (C) 2021 Even Rouault

""" This scripts merges the GRIB2 tables from DEGRIB, which are in the
    frmts/grib/degrib/data directory with the ones from WMO.
"""

import csv
import os
import re
import shutil

import requests

# URL of the WMO GRIB2 tables repository
wmo_github_repo = "wmo-im/GRIB2"
wmo_github_tag = "v34"
wmo_base_url = f"https://raw.githubusercontent.com/{wmo_github_repo}/{wmo_github_tag}"

script_dir_name = os.path.dirname(os.path.realpath(__file__))

# Input directory with degrib tables
degrib_table_input_dir = os.path.join(script_dir_name, "data")
assert os.path.exists(degrib_table_input_dir)

# Output directory of merged tables (frmts/grib/data)
output_dir = os.path.join(os.path.dirname(script_dir_name), "data")
assert os.path.exists(output_dir)

with open(os.path.join(output_dir, "grib2_table_versions.csv"), "wt") as f:
    f.write("component,version\n")
    f.write(f"wmo,{wmo_github_tag}\n")
    f.write("degrib,2.25\n")

r = requests.get(
    f"https://api.github.com/repos/{wmo_github_repo}/commits/HEAD",
    headers={"Accept": "application/vnd.github.VERSION.sha"},
)
wmo_table_sha1 = r.content.decode("utf-8")
warning_msg = "DO NOT MODIFY THIS FILE. It is generated by frmts/grib/degrib/merge_degrib_and_wmo_tables.py"
warning2_msg = (
    f"from tables at version https://github.com/wmo-im/GRIB2/commit/{wmo_table_sha1}"
)
warning_len = max(len(warning_msg), len(warning2_msg))

# Processes table 4.5 "Fixed surface types and units"


def process_table_4_5():

    # Ingest table from degrib
    degrib_csv_reader = csv.DictReader(
        open(os.path.join(degrib_table_input_dir, "grib2_table_4_5.csv"), newline="")
    )
    assert degrib_csv_reader.fieldnames == ["code", "short_name", "name", "unit"]
    records = [None for i in range(256)]
    for row in degrib_csv_reader:
        code = int(row["code"])
        assert records[code] is None
        row["code"] = code
        records[code] = row

    # Ingest table from WMO
    r = requests.get(wmo_base_url + "/GRIB2_CodeFlag_4_5_CodeTable_en.csv")
    lines = r.content.decode("utf-8").split("\n")
    wmo_csv_reader = csv.DictReader(lines)
    assert wmo_csv_reader.fieldnames == [
        "Title_en",
        "SubTitle_en",
        "CodeFlag",
        "Value",
        "MeaningParameterDescription_en",
        "Note_en",
        "noteIDs",
        "UnitComments_en",
        "Status",
    ], wmo_csv_reader.fieldnames
    wmo_records = [None for i in range(256)]
    for row in wmo_csv_reader:
        code = row["CodeFlag"]
        if "-" in code:
            # Deal with ranges (reserved codes)
            min_val, max_val = code.split("-")
            min_val = int(min_val)
            max_val = int(max_val)
            assert min_val < max_val
        else:
            min_val = int(code)
            max_val = int(code)
        for code in range(min_val, max_val + 1):
            assert wmo_records[code] is None
            wmo_records[code] = {
                "code": code,
                "name": row["MeaningParameterDescription_en"],
                "unit": row["UnitComments_en"],
            }

    # Merge records from WMO into degrib
    # degrib ones have the precedence when they exist since they contain
    # short_name.
    for code in range(len(records)):
        if wmo_records[code] is not None:
            if records[code] is None or (
                "Reserved" in records[code]["name"]
                and "Reserved" not in wmo_records[code]["name"]
            ):
                unit = wmo_records[code]["unit"]
                unit = "-" if unit == "" else unit
                records[code] = {
                    "code": code,
                    "short_name": "",
                    "name": wmo_records[code]["name"],
                    "unit": unit,
                }

    # Write the final file
    with open(
        os.path.join(output_dir, "grib2_table_4_5.csv"), "w", newline=""
    ) as csvfile:
        writer = csv.DictWriter(
            csvfile,
            fieldnames=degrib_csv_reader.fieldnames,
            quoting=csv.QUOTE_NONNUMERIC,
        )
        writer.writeheader()
        writer.writerow(
            {"code": -4, "short_name": "#" * warning_len, "name": "#", "unit": "#"}
        )
        writer.writerow(
            {"code": -3, "short_name": warning_msg, "name": "#", "unit": "#"}
        )
        writer.writerow(
            {"code": -2, "short_name": warning2_msg, "name": "#", "unit": "#"}
        )
        writer.writerow(
            {"code": -1, "short_name": "#" * warning_len, "name": "#", "unit": "#"}
        )
        for row in records:
            writer.writerow(row)


process_table_4_5()


# Process table referenced by 4.2 with "(Code table 4.X)" in the UnitComments_en
# column
def get_auxiliary_table(num):
    url = wmo_base_url + f"/GRIB2_CodeFlag_4_{num}_CodeTable_en.csv"
    print(f"Fetching {url}...")
    r = requests.get(url)
    lines = r.content.decode("utf-8").split("\n")
    wmo_csv_reader = csv.DictReader(lines)
    assert wmo_csv_reader.fieldnames == [
        "Title_en",
        "SubTitle_en",
        "CodeFlag",
        "Value",
        "MeaningParameterDescription_en",
        "Note_en",
        "noteIDs",
        "UnitComments_en",
        "Status",
    ], wmo_csv_reader.fieldnames
    ret = ""
    for row in wmo_csv_reader:
        if ret:
            ret += "; "
        ret += row["CodeFlag"]
        ret += "="
        ret += row["MeaningParameterDescription_en"]
    return ret


# Processes table 4.2 "Parameter Number by Product discipline and Parameter category"


def process_table_4_2():

    if os.path.exists("tmp_GRIB2"):
        shutil.rmtree("tmp_GRIB2")
    os.system(
        f"git clone --branch {wmo_github_tag} https://github.com/{wmo_github_repo} tmp_GRIB2"
    )
    lines = []
    # Merge together GRIB2_CodeFlag_4_2_xxxx.csv files
    for filename in os.listdir("tmp_GRIB2"):
        if filename.startswith("GRIB2_CodeFlag_4_2_"):
            new_lines = (
                open("tmp_GRIB2/" + filename, "rb").read().decode("utf-8").split("\n")
            )
            if len(lines) == 0:
                lines = new_lines
            else:
                lines += new_lines[1:]
    shutil.rmtree("tmp_GRIB2")

    wmo_csv_reader = csv.DictReader(lines)
    assert wmo_csv_reader.fieldnames == [
        "Title_en",
        "SubTitle_en",
        "CodeFlag",
        "Value",
        "MeaningParameterDescription_en",
        "Note_en",
        "noteIDs",
        "UnitComments_en",
        "Status",
    ], wmo_csv_reader.fieldnames

    # dict_tables has as key a subtable (by product discipline and category)
    # and as value a table of 256 entries with the subcategory as index
    dict_tables = {}

    auxiliary_tables = {}

    for row in wmo_csv_reader:

        # Split values
        # "Product discipline 0 - Meteorological products, parameter category 1: moisture""
        # as product_type = 0 and category = 1
        m = re.match(r".+?([0-9]+).+?([0-9]+).+", row["SubTitle_en"])
        product_type = int(m.group(1))
        category = int(m.group(2))

        # Instantiate final table
        table_name = f"grib2_table_4_2_{product_type}_{category}.csv"
        if table_name not in dict_tables:
            dict_tables[table_name] = [None for i in range(256)]
        table = dict_tables[table_name]

        code = row["CodeFlag"]

        if "-" in code:
            min_val, max_val = code.split("-")
            min_val = int(min_val)
            max_val = int(max_val)
            assert min_val < max_val
        else:
            min_val = int(code)
            max_val = int(code)

        name = row["MeaningParameterDescription_en"]
        unit = row["UnitComments_en"]
        m = re.match(r"\(Code table 4\.([0-9]+)\)", unit)
        if m is not None:
            auxiliary_table_num = m.group(1)
            if auxiliary_table_num not in auxiliary_tables:
                auxiliary_tables[auxiliary_table_num] = get_auxiliary_table(
                    auxiliary_table_num
                )
            unit = auxiliary_tables[auxiliary_table_num]

        for code in range(min_val, max_val + 1):
            assert table[code] is None, (table_name, code, table[code], row)
            table[code] = {
                "subcat": code,
                "short_name": "",
                "name": name,
                "unit": unit,
                "unit_conv": "UC_NONE",
                "override_unit": True if m is not None else False,
            }

    fieldnames = ["subcat", "short_name", "name", "unit", "unit_conv"]

    # Iterate over all sub tables.
    for filename, wmo_table in dict_tables.items():
        degrib_filename = os.path.join(degrib_table_input_dir, filename)

        # Ingest corresponding degrib subtable, if it exists
        if os.path.exists(degrib_filename):
            degrib_csv_reader = csv.DictReader(open(degrib_filename, newline=""))
            assert degrib_csv_reader.fieldnames == fieldnames
            records = [None for i in range(256)]
            for row in degrib_csv_reader:
                code = int(row["subcat"])
                assert records[code] is None
                row["subcat"] = code
                records[code] = row

            # Merge WMO and degrib subtable
            # degrib entries have precedence, unless there are reserved ranges
            # and WMO entry is not reserved
            for code in range(len(records)):
                if wmo_table[code] is not None:
                    if records[code] is None:
                        records[code] = wmo_table[code]
                    elif (
                        "Reserved" in records[code]["name"]
                        and "Reserved" not in wmo_table[code]["name"]
                    ):
                        # print('Patching ', records[code], ' with ', wmo_table[code])
                        records[code] = wmo_table[code]
                    if wmo_table[code]["override_unit"]:
                        records[code]["unit"] = wmo_table[code]["unit"]

        else:
            records = wmo_table

        # Write final subtable
        with open(os.path.join(output_dir, filename), "w", newline="") as csvfile:
            writer = csv.DictWriter(
                csvfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
            )
            writer.writeheader()
            writer.writerow(
                {
                    "subcat": -4,
                    "short_name": "#" * warning_len,
                    "name": "#",
                    "unit": "#",
                    "unit_conv": "#",
                }
            )
            writer.writerow(
                {
                    "subcat": -3,
                    "short_name": warning_msg,
                    "name": "#",
                    "unit": "#",
                    "unit_conv": "#",
                }
            )
            writer.writerow(
                {
                    "subcat": -2,
                    "short_name": warning2_msg,
                    "name": "#",
                    "unit": "#",
                    "unit_conv": "#",
                }
            )
            writer.writerow(
                {
                    "subcat": -1,
                    "short_name": "#" * warning_len,
                    "name": "#",
                    "unit": "#",
                    "unit_conv": "#",
                }
            )
            for row in records:
                if row:
                    writer.writerow({x: row[x] for x in fieldnames})


process_table_4_2()



#!/usr/bin/python

import os
import sys

if len(sys.argv) == 1:
    VERSION = open("VERSION", "rt").read()
    VERSION = VERSION.split(".")
    last_release_branch = "%d.%d" % (int(VERSION[0]), int(VERSION[1]) - 1)
else:
    last_release_branch = sys.argv[1]

print("Generating /tmp/log.txt with changes not in %s branch" % last_release_branch)

os.system(
    f'git log --reverse -v v{last_release_branch}.0..HEAD . ":(exclude)autotest" ":(exclude)doc" ":(exclude).github" > /tmp/log_raw.txt'
)
os.system(
    f'git log --no-merges --reverse -v v{last_release_branch}.0..release/{last_release_branch} . ":(exclude)autotest" ":(exclude)doc" ":(exclude).github" > /tmp/log_bugfixes.txt'
)


class Commit:
    def __init__(self):
        self.metadata = ""
        self.message = ""


def get_commits(filename):
    commits = []
    commit = None
    for l in open(filename, "rt").readlines():
        if l.startswith("commit "):
            commit = Commit()
            commits.append(commit)
            commit.metadata += l
        elif commit.message == "" and l == "\n":
            commit.message += l
        elif commit.message == "":
            commit.metadata += l
        else:
            commit.message += l
    return commits


raw_commits = get_commits("/tmp/log_raw.txt")
bugfixes_commits = get_commits("/tmp/log_bugfixes.txt")
set_bugfixes_commit_messages = set()
for commit in bugfixes_commits:
    set_bugfixes_commit_messages.add(commit.message)

with open("/tmp/log.txt", "wt") as f:
    for commit in raw_commits:
        is_bugfix = False
        message = commit.message.split("\n")[1]
        for bugfix_commit_message in set_bugfixes_commit_messages:
            if message in bugfix_commit_message:
                is_bugfix = True
                break
        if not is_bugfix:
            f.write(commit.metadata)
            f.write(commit.message)



#!/usr/bin/env python
# SPDX-License-Identifier: MIT
# Copyright (C) 2024 Even Rouault

import glob
import os

embedded_resources_gen1 = open(
    os.path.join(os.path.dirname(__file__), "embedded_resources_gen1.c"), "wt"
)
embedded_resources_gen2 = open(
    os.path.join(os.path.dirname(__file__), "embedded_resources_gen2.c"), "wt"
)

embedded_resources_gen1.write(
    "// File generated by generate_embedded_resources.py. DO NOT EDIT\n"
)
embedded_resources_gen2.write(
    "// File generated by generate_embedded_resources.py. DO NOT EDIT\n"
)

files = list(glob.glob(os.path.join(os.path.dirname(__file__), "data/*.gfs")))
files += list(glob.glob(os.path.join(os.path.dirname(__file__), "data/*.xml")))

for f in sorted(files):
    f = os.path.basename(f)
    c = f.replace(".", "_")
    embedded_resources_gen1.write(
        """
static const char %s [] = {
    #embed "data/%s"
            , 0
};
"""
        % (c, f)
    )
    embedded_resources_gen2.write("""{ "%s", %s },\n""" % (f, c))

embedded_resources_gen1.close()
embedded_resources_gen2.close()



import ast
import doctest
import functools

import numpy as np
import pytest

from osgeo import gdal, ogr, osr


###
# Make modules available to doctests
#
# Adding modules to doctest_namespace allows us to avoid writing "from osgeo import gdal"
# in every example.
###
@pytest.fixture(scope="session", autouse=True)
def global_setup(doctest_namespace):
    doctest_namespace["gdal"] = gdal
    doctest_namespace["ogr"] = ogr
    doctest_namespace["osr"] = osr
    doctest_namespace["np"] = np
    doctest_namespace["pytest"] = pytest


###
# Custom output checker
#
# doctest relies on an OutputChecker object to determine if test outputs
# match their expected values. pytest has its own OutputChecker implementation
# but does not provide any hooks for further customizations. So we inject a
# custom checker directly into pytest's internals.
###

import _pytest.doctest

base_checker = _pytest.doctest._get_checker()


class GDALOutputChecker(doctest.OutputChecker):
    @staticmethod
    def _strip_comments(want):
        lines = want.split("\n")
        for i in range(len(lines)):
            comment_pos = lines[i].find("#")
            if comment_pos != -1:
                lines[i] = lines[i][:comment_pos]
        return "\n".join(lines)

    @staticmethod
    def _strip_warnings(want):
        lines = want.split("\n")
        return "\n".join(x for x in lines if not x.startswith("Warning"))

    def check_output(self, want, got, flags):
        want = self._strip_warnings(want)

        # custom checking method for arrays
        # pytest's built-in NUMBER flag is not suitable because
        # it only allows differences to the right of the decimal
        # point.
        if "rtol:" in want:
            pos = want.find("rtol:")
            rtol = float(want[(pos + 5) :].strip())
            want = np.array(ast.literal_eval(want))
            got = np.array(ast.literal_eval(got))
            return np.allclose(want, got, rtol=rtol)

        if "# random" in want:
            want = np.array(ast.literal_eval(want))
            got = np.array(ast.literal_eval(got))
            return np.array_equal(want.shape, got.shape)

        if "# no-check" in want:
            return True

        want = self._strip_comments(want)

        return base_checker.check_output(want, got, flags)


_pytest.doctest.CHECKER_CLASS = GDALOutputChecker

###
# Path substitutions
#
# We want to write concise documentation using short filenames like gdal.Open("byte.tif")
# Modifying the working directory before each doctest is tedious and error-prone.
# Instead we store a list of abbreviated filenames and monkey-patch GDAL functions to
# substitute the full path name before running the example.
###

files = {
    "byte.tif": "autotest/gcore/data/byte.tif",
    "my.tif": "autotest/gcore/data/byte.tif",
    "in.tif": "autotest/gcore/data/byte.tif",
    "poly.shp": "autotest/ogr/data/poly.shp",
    "testrat.tif": "autotest/gcore/data/gtiff/testrat.tif",
}


def nested_replace(x):
    if type(x) is int or type(x) is bool:
        return x
    if type(x) is str:
        return files.get(x, x)
    if type(x) is dict:
        return {k: nested_replace(v) for k, v in x.items()}
    if type(x) is tuple:
        return (nested_replace(v) for v in x)
    if type(x) is list:
        return [nested_replace(v) for v in x]
    if isinstance(x, (gdal.Algorithm, gdal.Dataset)):
        return x

    raise Exception(f"Unsupported argument type: {type(x)}")


def expand_dsn(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*nested_replace(args), **nested_replace(kwargs))

    return wrapper


@pytest.fixture(autouse=True)
def gdal_find_files(monkeypatch):

    monkeypatch.setattr(gdal, "Open", expand_dsn(gdal.Open))
    monkeypatch.setattr(gdal, "OpenEx", expand_dsn(gdal.OpenEx))
    monkeypatch.setattr(gdal, "Run", expand_dsn(gdal.Run))
    monkeypatch.setattr(
        gdal.Algorithm,
        "ParseCommandLineArguments",
        expand_dsn(gdal.Algorithm.ParseCommandLineArguments),
    )
    monkeypatch.setattr(
        gdal.Algorithm, "__setitem__", expand_dsn(gdal.Algorithm.__setitem__)
    )



#!/usr/bin/python
# Check that driver documentation pages are consistent with driver capabilities

import os

from osgeo import gdal

doc_source = os.path.join(os.path.dirname(os.path.dirname(__file__)), "doc", "source")

map_doc_caps = {}

for subdir in ("raster", "vector"):
    dirname = os.path.join(doc_source, "drivers", subdir)
    for f in os.listdir(dirname):
        filename = os.path.join(dirname, f)
        if os.path.isfile(filename):
            shortnames = []
            supports_create = False
            supports_createcopy = False
            for l in open(filename, "rt").readlines():
                if l.startswith(".. shortname:: "):
                    shortnames.append(l[len(".. shortname:: ") : -1])
                elif "supports_create::" in l:
                    supports_create = True
                elif "supports_createcopy::" in l:
                    supports_createcopy = True
            for shortname in shortnames:
                d = {}
                if supports_create:
                    d["supports_create"] = True
                if supports_createcopy:
                    d["supports_createcopy"] = True
                if shortname.upper() in map_doc_caps:
                    assert subdir == "vector"
                    map_doc_caps["OGR_" + shortname.upper()] = d
                else:
                    map_doc_caps[shortname.upper()] = d

for i in range(gdal.GetDriverCount()):
    drv = gdal.GetDriver(i)
    shortname = drv.ShortName
    if shortname in (
        "BIGGIF",
        "GTX",
        "NULL",
        "GNMFile",
        "GNMDatabase",
        "HTTP",
        "AVIF_HEIF",
    ):
        continue
    if shortname == "OGR_GMT":
        shortname = "GMT"
    if shortname == "OGR_OGDI":
        shortname = "OGDI"
    if shortname == "NWT_GRC":
        # mixed in same page as NWT_GRD, which confuses our logic
        continue
    if shortname == "SQLite":
        # sqlite.rst and rasterlite2.rst declare both SQLite driver name,
        # which confuses our logic
        continue
    if shortname in ("OpenFileGDB", "NGW"):
        # one page in vector and another one in raster, which confuses our logic
        continue

    assert shortname.upper() in map_doc_caps, drv.ShortName
    doc_caps = map_doc_caps[shortname.upper()]

    if drv.GetMetadataItem(gdal.DCAP_CREATE):
        if shortname == "PDF":
            continue  # Supports Create() but in a very specific mode, hence better not advertizing it
        assert (
            "supports_create" in doc_caps
        ), f"Driver {shortname} declares DCAP_CREATE but doc does not!"
    else:
        if shortname == "HDF4":
            continue  # This is actually the HDF4Image
        assert (
            "supports_create" not in doc_caps
        ), f"Driver {shortname} does not declare DCAP_CREATE but doc does!"

    if drv.GetMetadataItem(gdal.DCAP_CREATECOPY):
        if shortname == "WMTS":
            continue  # Supports CreateCopy() but in a very specific mode, hence better not advertizing it
        assert (
            "supports_createcopy" in doc_caps
        ), f"Driver {shortname} declares DCAP_CREATECOPY but doc does not!"
    else:
        if not drv.GetMetadataItem(gdal.DCAP_CREATE):
            if shortname in ("AVIF", "JP2MrSID"):
                continue  # build dependent
            assert (
                "supports_createcopy" not in doc_caps
            ), f"Driver {shortname} does not declare DCAP_CREATECOPY but doc does!"



#!/usr/bin/env python3

import sys

from osgeo.gdal import deprecation_warn

# import osgeo_utils.gdal_edit as a convenience to use as a script
from osgeo_utils.gdal_edit import *  # noqa
from osgeo_utils.gdal_edit import main

deprecation_warn("gdal_edit")
sys.exit(main(sys.argv))



#!/usr/bin/env python3

"""This file checks that the root directory does not contain an unexpected binary file"""

import csv
import os
import subprocess
import sys

this_dir = os.path.dirname(__file__)
if not os.path.isabs(this_dir):
    this_dir = os.path.join(os.getcwd(), this_dir)
this_dir = os.path.abspath(this_dir)
root_dir = os.path.dirname(this_dir)

binaries_allow_list = {}
with open(
    os.path.join(this_dir, "binaries_allow_list.csv"), newline="", encoding="utf-8"
) as f:
    reader = csv.DictReader(f)
    assert reader.fieldnames == ["filename", "sha256sum"]
    for row in reader:
        binaries_allow_list[row["filename"]] = row["sha256sum"]

error_code = 0

for dirname in os.listdir(root_dir):
    # We skip doc and autotest as they are not included in the GDAL source tarball
    if dirname not in ("doc", "autotest", ".git"):
        p = subprocess.Popen(
            [
                "find",
                os.path.join(root_dir, dirname),
                "-type",
                "f",
                "-exec",
                "file",
                "{}",
                ";",
            ],
            stdout=subprocess.PIPE,
        )
        out, _ = p.communicate()
        for line in out.decode("utf-8").split("\n"):
            if not line:
                continue
            tokens = line.split(":")
            filename = tokens[0]
            kind = ":".join(tokens[1:]).strip()
            if (
                "text" not in kind
                and "AutoCAD" not in kind
                and kind not in ("empty", "JSON data")
            ):
                p = subprocess.Popen(["sha256sum", filename], stdout=subprocess.PIPE)
                sha256sum, _ = p.communicate()
                sha256sum = sha256sum.decode("utf-8").split("  ")[0]

                rel_filename = filename[len(root_dir) + 1 :]
                if rel_filename not in binaries_allow_list:
                    error_code = 1
                    print(
                        f'Found unknown binary file {rel_filename} of kind "{kind}". If it is legit, add the following line in scripts/binaries_allow_list.csv:\n"{rel_filename}",{sha256sum}'
                    )
                elif binaries_allow_list[rel_filename] != sha256sum:
                    error_code = 1
                    print(
                        f'Binary file {rel_filename} has a different sha256sum than expected. If it is legit, update the following line in scripts/binaries_allow_list.csv:\n"{rel_filename}",{sha256sum}'
                    )

sys.exit(error_code)



#!/usr/bin/env python3

# This script must ideally be run on a configuration that has all drivers enabled...

import glob
import os
import sys

prefixes = {}
prefixes["GTI:"] = "GTI"
prefixes["GMLAS:"] = "GMLAS"
prefixes["HANA:"] = "HANA"
prefixes["PG:"] = "PostgreSQL"
prefixes["MSSQL:"] = "MSSQLSpatial"
prefixes["ES:"] = "Elasticsearch"


def collect_prefix(filename):
    lines = open(filename, "rt").readlines()
    i = -1
    len_lines = len(lines)

    while True:
        i += 1
        if i >= len_lines:
            break
        l = lines[i][0:-1].strip()
        if l.startswith("/*"):
            continue

        pos = l.find('SetMetadataItem(GDAL_DMD_CONNECTION_PREFIX, "')
        if pos > 0:
            pos += len('SetMetadataItem(GDAL_DMD_CONNECTION_PREFIX, "')
            pos_end = l.find('")', pos)
            assert pos_end > 0
            prefix = l[pos:pos_end]
            assert prefix.endswith(":"), prefix
            if prefix not in prefixes:
                prefixes[prefix] = prefix[:-1]


def explore(dirname):
    for filename in glob.glob(dirname + "/*"):
        if os.path.isdir(filename):
            explore(filename)
        if filename.endswith(".cpp"):
            collect_prefix(filename)


for dirname in ("frmts", "ogr"):
    explore(os.path.join(os.path.dirname(__file__), "..", dirname))

ret = 0

has_gdal = False
try:
    from osgeo import gdal

    has_gdal = True
except ImportError:
    pass

if has_gdal:
    # Do a runtime check to verify that there is no missing prefix through source
    # code exploration.
    for i in range(gdal.GetDriverCount()):
        drv = gdal.GetDriver(i)
        connection_prefix = drv.GetMetadataItem(gdal.DMD_CONNECTION_PREFIX)
        if connection_prefix and connection_prefix not in prefixes:
            print(
                "Driver %s report prefix %s, but not found through source code discovery. Manually add it at top of collect_driver_connection_prefix.py"
                % (drv.ShortName, connection_prefix)
            )
            ret = 1
        elif (
            connection_prefix
            and connection_prefix in prefixes
            and prefixes[connection_prefix].upper() != drv.ShortName.upper()
            and drv.ShortName != "PostGISRaster"
        ):
            print(
                "Mismatch between prefix %s and driver name for %s. Manually add it at top of collect_driver_connection_prefix.py"
                % (connection_prefix, drv.ShortName)
            )
            ret = 1

c_file = (
    "// This file was automatically generated by collect_driver_connection_prefix.py\n"
)
c_file += "// DO NOT EDIT\n\n"
c_file += "// clang-format off\n"
c_file += "constexpr static struct {\n"
c_file += "  const char* pszPrefix;\n"
c_file += "  const char* pszDriverName;\n"
c_file += "}\n"
c_file += "asKnownConnectionPrefixes[] =\n"
c_file += "{\n"
for prefix in sorted(prefixes, key=str.lower):
    c_file += f'   {{"{prefix}", "{prefixes[prefix]}"}},\n'
c_file += "};\n"
c_file += "\n"
c_file += "// clang-format on\n"
# print(c_file)

open(
    os.path.join(
        os.path.dirname(__file__), "..", "gcore", "gdal_known_connection_prefixes.h"
    ),
    "wt",
).write(c_file)

sys.exit(ret)



#!/usr/bin/env python
###############################################################################
#  Project:  GDAL samples
#  Purpose:  Replace container.size() by !container.empty(), and reverse
#  Author:   Even Rouault <even.rouault at spatialys.com>
#
###############################################################################
#  Copyright (c) 2016, Even Rouault <even.rouault at spatialys.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
###############################################################################


import sys


def find_start_identifier_pos(content, pos_start_identifier):
    level_parenthesis = 0
    level_bracket = 0
    while True:
        c = content[pos_start_identifier]
        if level_parenthesis > 0 and c != "(" and c != ")":
            pos_start_identifier -= 1
        elif c == ")":
            level_parenthesis += 1
            pos_start_identifier -= 1
        elif c == "(":
            if level_parenthesis == 0:
                break
            level_parenthesis -= 1
            pos_start_identifier -= 1
        elif level_bracket > 0 and c != "[" and c != "]":
            pos_start_identifier -= 1
        elif c == "]":
            level_bracket += 1
            pos_start_identifier -= 1
        elif c == "[":
            if level_bracket == 0:
                break
            level_bracket -= 1
            pos_start_identifier -= 1
        elif c.isalnum() or c == "_" or c == ".":
            pos_start_identifier -= 1
        elif c == ">" and content[pos_start_identifier - 1] == "-":
            pos_start_identifier -= 2
        else:
            break
    pos_start_identifier += 1
    return pos_start_identifier


content = open(sys.argv[1], "rb").read()
pos = 0
modified = False
while True:
    pos1 = content.find(".size()", pos)
    pos2 = content.find("->size()", pos)
    if pos1 < 0 and pos2 < 0:
        break
    separator = ""
    if pos1 >= 0 and (pos1 < pos2 or pos2 < 0):
        pos = pos1
        pos_after = pos + len(".size()")
        separator = "."
    else:
        pos = pos2
        dot_variant = False
        pos_after = pos + len("->size()")
        separator = "->"

    if content[pos_after] == " ":
        pos_after += 1
    extra_space = ""
    empty = False
    non_empty = False
    if content[pos_after:].startswith("== 0"):
        empty = True
        pos_after += len("== 0")
    elif content[pos_after:].startswith("==0"):
        empty = True
        pos_after += len("==0")
    elif content[pos_after:].startswith("!= 0"):
        non_empty = True
        pos_after += len("!= 0")
    elif content[pos_after:].startswith("!=0"):
        non_empty = True
        pos_after += len("!=0")
    elif content[pos_after:].startswith("> 0"):
        non_empty = True
        pos_after += len("> 0")
    elif content[pos_after:].startswith(">0"):
        non_empty = True
        pos_after += len(">0")

    if (
        not empty
        and not non_empty
        and (
            content[pos_after:].startswith(" )")
            or content[pos_after:].startswith(")")
            or content[pos_after:].startswith(" &&")
            or content[pos_after:].startswith(" ||")
            or content[pos_after:].startswith("&&")
            or content[pos_after:].startswith("||")
            or content[pos_after:].startswith(" ?")
            or content[pos_after:].startswith("?")
        )
    ):
        if content[pos_after] != " ":
            extra_space = " "
        pos_cur = find_start_identifier_pos(content, pos - 1)
        pos_cur -= 1
        while content[pos_cur] in [" ", "\n", "\t"]:
            pos_cur -= 1
        pos_cur += 1
        if (
            content[pos_after:].startswith(" ?") or content[pos_after:].startswith("?")
        ) and (content[pos_cur - 1] == "(" or content[pos_cur - 1] == ","):
            non_empty = True
        elif content[pos_cur - 3 : pos_cur] == "if(":
            non_empty = True
        elif content[pos_cur - 4 : pos_cur] == "if (":
            non_empty = True
        elif content[pos_cur - 2 : pos_cur] == "&&":
            non_empty = True
        elif content[pos_cur - 2 : pos_cur] == "||":
            non_empty = True

    if empty:
        modified = True
        content = content[0:pos] + separator + "empty()" + content[pos_after:]
    elif non_empty:
        modified = True
        pos_start_identifier = find_start_identifier_pos(content, pos - 1)
        content = (
            content[0:pos_start_identifier]
            + "!"
            + content[pos_start_identifier:pos]
            + separator
            + "empty()"
            + extra_space
            + content[pos_after:]
        )

    pos += 1

if modified:
    open(sys.argv[1], "wb").write(content)

# sys.stdout.write(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# Project:  GDAL/OGR Utilities
# Purpose:  Bash completion script generation
# Author:   Guillaume Pasero <guillaume dot pasero at c dash s dot fr>
#
###############################################################################
# Copyright (c) 2016, Guillaume Pasero <guillaume dot pasero at c dash s dot fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################

import sys
from subprocess import PIPE, STDOUT, Popen


def showHelp():
    print("Usage : completionFinder.py  output_script")


def processLine(outList, line):

    in_option = False
    item = ""
    last_c = ""
    for c in line:
        if last_c in ("[", "|") and c == "-" and not in_option:
            in_option = True
            item = c
        elif in_option:
            if c not in ("]", "[", " ", "|"):
                item += c
            else:
                in_option = False
                key = item
                item = ""
                # Handle special cases -key={A/B/C}
                if key.count("={") == 1 and key[-1] == "}":
                    pair = key.split("=")
                    choices = ((pair[1])[1:-1]).split("/")
                    for c in choices:
                        outList.append(pair[0] + "=" + c)
                else:
                    pos = key.find("=")
                    if pos > 0:
                        key = key[0:pos]
                    if key != "-" and key not in outList:
                        outList.append(key)
        last_c = c

    return outList


def processTool(toolName):
    if toolName == "gdal_calc.py":
        command = [toolName]
    else:
        command = [toolName, "--help"]

    process = Popen(command, stdout=PIPE, stderr=STDOUT)
    result = process.communicate()[0]
    lines = result.decode("utf-8").split("\n")
    options = []

    for line in lines:
        cleanLine = line.strip("\t\r ")
        options = processLine(options, cleanLine)

    return options


def parseGDALGeneralOptions():
    command = ["gdalinfo", "--help-general"]
    process = Popen(command, stdout=PIPE, stderr=STDOUT)
    result = process.communicate()[0]
    lines = result.decode("utf-8").split("\n")
    index = 0
    options = []
    while index < len(lines):
        cleanLine = lines[index].strip("\t\r ")
        parts = cleanLine.split(":")
        if parts[0].find("--") == 0:
            words = parts[0].split(" ")
            options.append(words[0])
        index += 1
    return options


def parseOGRGeneralOptions():
    command = ["ogr2ogr", "--help-general"]
    process = Popen(command, stdout=PIPE, stderr=STDOUT)
    result = process.communicate()[0]
    lines = result.decode("utf-8").split("\n")
    index = 0
    options = []
    while index < len(lines):
        cleanLine = lines[index].strip("\t\r ")
        parts = cleanLine.split(":")
        if parts[0].find("--") == 0:
            words = parts[0].split(" ")
            options.append(words[0])
        index += 1
    return options


# generate completion script section for :
#   - name : the given gdal/ogr tool
#   - optList : option list


def getCompletionScript(name, optList):
    output = []

    if name.endswith(".py"):
        func_name = name[0:-3]
    else:
        func_name = name

    output.append("_" + func_name + "()\n")
    output.append("{\n")
    output.append("  COMPREPLY=()\n")
    output.append('  cur="${COMP_WORDS[$COMP_CWORD]}"\n')
    output.append('  case "$cur" in\n')
    output.append("    -*)\n")
    optLine = '      key_list="'
    for item in optList:
        optLine += item + " "
    optLine += '"\n'
    output.append(optLine)
    output.append('      if [ "$CURRENT_SHELL" = "bash" ]; then\n')
    output.append(
        '        mapfile -t COMPREPLY < <(compgen -W "$key_list" -- "$cur")\n'
    )
    output.append("      else\n")
    output.append('        COMPREPLY=( "$(compgen -W "$key_list" -- "$cur")" )\n')
    output.append("      fi\n")
    output.append("      return 0\n")
    output.append("      ;;\n")
    output.append("  esac\n")

    # adapt format parsing command : GDAL type of OGR type

    isGdal = True
    if name.startswith("ogr"):
        isGdal = False

    # gdal type
    if isGdal:
        formatParsingCmd = "$tool --formats | tail -n +2 | cut -f 3 -d ' '"
        if ("-ot" in optList) or ("-of" in optList) or ("--format" in optList):
            output.append("  tool=${COMP_WORDS[0]}\n")
            output.append('  prev="${COMP_WORDS[$COMP_CWORD-1]}"\n')
            output.append('  case "$prev" in\n')
            if "-ot" in optList:
                output.append("    -ot)\n")
                output.append(
                    '      key_list="Byte Int16 UInt16 UInt32 Int32 Float32 Float64 CInt16 CInt32 CFloat32 CFloat64"\n'
                )
                output.append('      if [ "$CURRENT_SHELL" = "bash" ]; then\n')
                output.append(
                    '        mapfile -t COMPREPLY < <(compgen -W "$key_list" -- "$cur")\n'
                )
                output.append("      else\n")
                output.append(
                    '        COMPREPLY=( "$(compgen -W "$key_list" -- "$cur")" )\n'
                )
                output.append("      fi\n")
                output.append("      ;;\n")
            for arg in ("-f", "-of"):
                if (arg in optList) and isGdal:
                    output.append("    %s)\n" % arg)
                    output.append('      key_list="$( ' + formatParsingCmd + ')"\n')
                    output.append('      if [ "$CURRENT_SHELL" = "bash" ]; then\n')
                    output.append(
                        '        mapfile -t COMPREPLY < <(compgen -W "$key_list" -- "$cur")\n'
                    )
                    output.append("      else\n")
                    output.append(
                        '        COMPREPLY=( "$(compgen -W "$key_list" -- "$cur")" )\n'
                    )
                    output.append("      fi\n")
                    output.append("      ;;\n")
            if ("--format" in optList) and isGdal:
                output.append("    --format)\n")
                output.append('      key_list="$( ' + formatParsingCmd + ')"\n')
                output.append('      if [ "$CURRENT_SHELL" = "bash" ]; then\n')
                output.append(
                    '        mapfile -t COMPREPLY < <(compgen -W "$key_list" -- "$cur")\n'
                )
                output.append("      else\n")
                output.append(
                    '        COMPREPLY=( "$(compgen -W "$key_list" -- "$cur")" )\n'
                )
                output.append("      fi\n")
                output.append("      ;;\n")
            output.append("  esac\n")
    else:
        # ogr type
        formatParsingCmd = (
            "$tool --formats | tail -n +2 | grep -o -E '\"[^\"]+\"' | sed 's/\\ /__/'"
        )
        if "-f" in optList:
            # replace ogrtindex by ogr2ogr to check --formats
            output.append("  tool=${COMP_WORDS[0]/ogrtindex/ogr2ogr}\n")
            output.append('  prev="${COMP_WORDS[$COMP_CWORD-1]}"\n')
            output.append('  case "$prev" in\n')
            for arg in ("-f", "-of"):
                if (arg in optList) and not isGdal:
                    # completion is more tricky here because of spaces
                    output.append("    %s)\n" % arg)
                    output.append('      key_list="$( ' + formatParsingCmd + ')"\n')
                    output.append("      for iter in $key_list; do\n")
                    output.append("        if [[ $iter =~ ^$cur ]]; then\n")
                    output.append('          COMPREPLY+=( "${iter//__/ }" )\n')
                    output.append("        fi\n")
                    output.append("      done\n")
                    output.append("      ;;\n")
            output.append("  esac\n")

    output.append("  return 0\n")
    output.append("}\n")
    output.append("complete -o default -F _" + func_name + " " + name + "\n")
    if name.endswith(".py"):
        output.append("complete -o default -F _" + func_name + " " + name[0:-3] + "\n")

    return output


def main(argv):
    if len(argv) < 2:
        showHelp()
        return 1

    gdaltools = [
        "gdal2tiles.py",
        "gdal2xyz.py",
        "gdaladdo",
        "gdalbuildvrt",
        "gdal_calc.py",
        "gdalcompare.py",
        "gdal-config",
        "gdal_contour",
        "gdaldem",
        "gdal_edit.py",
        "gdalenhance",
        "gdal_fillnodata.py",
        "gdal_grid",
        "gdalinfo",
        "gdallocationinfo",
        "gdalmanage",
        "gdal_merge.py",
        "gdalmove.py",
        "gdal_polygonize.py",
        "gdal_proximity.py",
        "gdal_rasterize",
        "gdal_retile.py",
        "gdal_sieve.py",
        "gdalsrsinfo",
        "gdaltindex",
        "gdaltransform",
        "gdal_translate",
        "gdalwarp",
        "gdal_viewshed",
        "gdal_create",
        "sozip",
        "gdal_footprint",
    ]

    ogrtools = [
        "ogr2ogr",
        "ogrinfo",
        "ogrlineref",
        "ogrtindex",
        "ogrmerge.py",
        "ogr_layer_algebra.py",
    ]

    # parse general options
    generalOptions = parseGDALGeneralOptions()
    generalOGROptions = parseOGRGeneralOptions()

    outFile = argv[1]
    of = open(outFile, "w")
    of.write("# shellcheck shell=bash disable=SC2148\n")
    of.write(
        "# WARNING: File auto-generated by completionFinder.py, do not modify manually\n"
    )
    of.write(
        """

if ! ps -p $$ >/dev/null 2>/dev/null; then
  # For busybox
  CURRENT_SHELL="bash"
else
  CURRENT_SHELL="$(sh -c 'ps -p $$ -o ppid=' | xargs ps -o comm= -p)"
fi

if [ "$CURRENT_SHELL" = "zsh" ]; then
  autoload -U bashcompinit
  bashcompinit
fi

HAS_GET_COMP_WORDS_BY_REF=no
if [ "$CURRENT_SHELL" = "bash" ]; then
  function_exists() {
      declare -f -F "$1" > /dev/null
      return $?
  }

  # Checks that bash-completion is recent enough
  if function_exists _get_comp_words_by_ref; then
    HAS_GET_COMP_WORDS_BY_REF=yes;
  fi
fi

_gdal()
{
  COMPREPLY=()
  cur="${COMP_WORDS[$COMP_CWORD]}"

  if test "$HAS_GET_COMP_WORDS_BY_REF" = "yes"; then
    local cur prev
    _get_comp_words_by_ref cur prev
    if test "$cur" = ""; then
      extra="last_word_is_complete=true"
    else
      extra="prev=${prev} cur=${cur}"
    fi
    choices=$(gdal completion ${COMP_LINE} ${extra})
  else
    choices=$(gdal completion ${COMP_LINE})
  fi
  if [ "$CURRENT_SHELL" = "bash" ]; then
    if [[ "$cur" == "=" ]]; then
      mapfile -t COMPREPLY < <(compgen -W "$choices" --)
    elif [[ "$cur" == ":" ]]; then
      mapfile -t COMPREPLY < <(compgen -W "$choices" --)
    elif [[ "$cur" == "!" ]]; then
      mapfile -t COMPREPLY < <(compgen -W "$choices" -P "! " --)
    else
      mapfile -t COMPREPLY < <(compgen -W "$choices" -- "$cur")
    fi
    for element in "${COMPREPLY[@]}"; do
      if [[ $element == */ ]]; then
        # Do not add a space if one of the suggestion ends with slash
        compopt -o nospace
        break
      elif [[ $element == *= ]]; then
        # Do not add a space if one of the suggestion ends with equal
        compopt -o nospace
        break
      elif [[ $element == *: ]]; then
        # Do not add a space if one of the suggestion ends with colon
        compopt -o nospace
        break
      fi
    done
  else
    # zsh
    if [ "$cur" = "=" ]; then
      COMPREPLY=( "$(compgen -W "$choices" --)" )
    elif [ "$cur" = ":" ]; then
      COMPREPLY=( "$(compgen -W "$choices" --)" )
    elif [ "$cur" = "!" ]; then
      COMPREPLY=( "$(compgen -W "$choices" -P "! " --)" )
    else
      COMPREPLY=( "$(compgen -W "$choices" -- "$cur")" )
    fi
    for element in "${COMPREPLY[@]}"; do
      if [[ $element == */ ]]; then
        # Do not add a space if one of the suggestion ends with slash
        _COMP_OPTIONS+=(nospace)
        break
      elif [[ $element == *= ]]; then
        # Do not add a space if one of the suggestion ends with equal
        _COMP_OPTIONS+=(nospace)
        break
      elif [[ $element == *: ]]; then
        # Do not add a space if one of the suggestion ends with colon
        _COMP_OPTIONS+=(nospace)
        break
      fi
    done
  fi
}

# nosort requires bash 4.4
complete -o nosort -o default -F _gdal gdal 2>/dev/null || complete -o default -F _gdal gdal

"""
    )

    for name in gdaltools:
        # verbose print
        print(name)
        optList = processTool(name)
        if "--help-general" in optList:
            for item in generalOptions:
                if item not in optList:
                    optList.append(item)
        of.writelines(getCompletionScript(name, optList))
    for name in ogrtools:
        # verbose print
        print(name)
        optList = processTool(name)
        if "--help-general" in optList:
            for item in generalOGROptions:
                if item not in optList:
                    optList.append(item)
        of.writelines(getCompletionScript(name, optList))

    of.close()
    return 0


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# Project:  GDAL/OGR
# Purpose:  Update copyright info in headers
# Author:   Even Rouault <even dot rouault at spatialys.com>
#
###############################################################################
# Copyright (c) 2014, Even Rouault <even dot rouault at spatialys.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################

import os

# WARNING: Only works from a git repository !

# Please edit for a different author
author_in_file = "rouault"
git_author = "Even Rouault"
full_author = "Even Rouault <even dot rouault at mines-paris dot org>"

for dirname, dirnames, filenames in os.walk("."):

    if ".svn" in dirname:
        continue
    if ".git" in dirname:
        continue
    if "libpng" in dirname:
        continue
    if "libjpeg" in dirname:
        continue
    if "libtiff" in dirname:
        continue
    if "giflib" in dirname:
        continue
    if "libgeotiff" in dirname:
        continue
    if "libjson" in dirname:
        continue

    # print path to all filenames.
    for filename in filenames:
        if ".svn" in filename:
            continue
        if ".h" not in filename and ".c" not in filename and ".py" not in filename:
            continue
        if filename == "e00read.c" or filename == "e00compr.h":
            continue
        fullfilename = os.path.join(dirname, filename)

        # print(fullfilename)
        minyear = -1
        maxyear = -1

        # Find if we have authored something and commit numbers to ignore (commits that originate from other authors)
        ret = os.popen("git log %s" % fullfilename)
        lines = ret.readlines()
        i = 0
        nlines = len(lines)
        found_commit = False
        commits_to_ignore = []
        while i < nlines:
            line = lines[i][0:-1]
            if line.startswith("Author:") and git_author in line:
                i = i + 1
                # line = lines[i][0:-1]
                # year = int(line.split(' ')[7])
                i = i + 1
                i = i + 1
                commit_number = lines[i - 4][7:15]
                ignore_commit = False
                while i < nlines:
                    line = lines[i][0:-1]
                    if line.startswith("commit"):
                        break
                    if (
                        "(by" in line
                        or "contributed by" in line
                        or "patch" in line
                        or "Patch" in line
                        or "Update copyright" in line
                        or "From: " in line
                        or "Add BLX Magellan Topo driver" in line
                    ):
                        ignore_commit = True
                    i = i + 1
                if ignore_commit:
                    commits_to_ignore.append(commit_number)
                else:
                    found_commit = True
            else:
                i = i + 1

        if not found_commit:
            continue

        # Count how many lines we have authored
        ret = os.popen("git blame %s" % fullfilename)
        lines = ret.readlines()
        i = 0
        nlines = len(lines)
        count_matching_lines = 0
        while i < nlines:
            line = lines[i][0:-1]
            ignore = False
            for commit in commits_to_ignore:
                if line.startswith(commit):
                    ignore = True
            if not ignore and "Copyright" not in line:
                idx = line.find("(" + git_author)
                if idx > 0:
                    count_matching_lines = count_matching_lines + 1
                    line = line[idx + len(git_author) + 1 :]
                    idx = line.find(" 20")
                    year = int(line[idx + 1 : idx + 5])
                    if minyear < 0 or year < minyear:
                        minyear = year
                    if maxyear < 0 or year > maxyear:
                        maxyear = year
            i = i + 1

        # Only claim copyright if we have authored more than 10 lines.
        if count_matching_lines < 10:
            continue

        print(fullfilename + " %d-%d" % (minyear, maxyear))
        f = open(fullfilename, "rb")
        f2 = open(fullfilename + ".tmp", "wb")
        lines = f.readlines()
        i = 0
        nlines = len(lines)
        already_added = False
        if ".py" in filename:
            prefix = "# "
        else:
            prefix = " * "
        while i < nlines:
            line = lines[i]
            if not already_added and "Copyright" in line:
                already_added = True
                if line.startswith("#  "):
                    prefix = "#  "
                elif line.startswith("# * "):
                    prefix = "# * "
                elif line.startswith("// "):
                    prefix = "// "
                while author_in_file not in line:
                    f2.write(line)
                    i = i + 1
                    line = lines[i]
                    if (
                        "Copyright" not in line and len(line.strip()) < 10
                    ) or line.find("Permission to use") > 0:
                        break
                if minyear < maxyear:
                    f2.write(
                        "%sCopyright (c) %d-%d, %s\n"
                        % (prefix, minyear, maxyear, full_author)
                    )
                else:
                    f2.write(
                        "%sCopyright (c) %d, %s\n" % (prefix, minyear, full_author)
                    )
                if author_in_file not in line:
                    f2.write(line)
            else:
                f2.write(line)
            i = i + 1
        f.close()
        f2.close()
        os.rename(fullfilename + ".tmp", fullfilename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is in the public domain, so as to serve as a template for
# real-world plugins.
# or, at the choice of the licensee,
# Copyright 2019 Even Rouault
# SPDX-License-Identifier: MIT

# Metadata parsed by GDAL C++ code at driver pre-loading, starting with '# gdal: '
# Required and with that exact syntax since it is parsed by non-Python
# aware code. So just literal values, no expressions, etc.
# gdal: DRIVER_NAME = "DUMMY"
# API version(s) supported. Must include 1 currently
# gdal: DRIVER_SUPPORTED_API_VERSION = [1]
# gdal: DRIVER_DCAP_VECTOR = "YES"
# gdal: DRIVER_DMD_LONGNAME = "my super plugin"

# Optional driver metadata items.
# # gdal: DRIVER_DMD_EXTENSIONS = "ext1 est2"
# # gdal: DRIVER_DMD_HELPTOPIC = "http://example.com/my_help.html"

try:
    # The gdal_python_driver module is defined by the GDAL library at runtime
    from gdal_python_driver import BaseDataset, BaseDriver, BaseLayer
except ImportError:
    # To be able to run in standalone mode
    class BaseDriver:
        pass

    class BaseDataset:
        pass

    class BaseLayer:
        pass


class Layer(BaseLayer):
    def __init__(self):

        # Reserved attribute names. Either those or the corresponding method
        # must be defined
        self.name = "my_layer"  # Required, or name() method

        self.fid_name = "my_fid"  # Optional

        self.fields = [
            {"name": "boolField", "type": "Boolean"},
            {"name": "int16Field", "type": "Integer16"},
            {"name": "int32Field", "type": "Integer"},
            {"name": "int64Field", "type": "Integer64"},
            {"name": "realField", "type": "Real"},
            {"name": "floatField", "type": "Float"},
            {"name": "strField", "type": "String"},
            {"name": "strNullField", "type": "String"},
            {"name": "strUnsetField", "type": "String"},
            {"name": "binaryField", "type": "Binary"},
            {"name": "timeField", "type": "Time"},
            {"name": "dateField", "type": "Date"},
            {"name": "datetimeField", "type": "DateTime"},
        ]  # Required, or fields() method

        self.geometry_fields = [
            {
                "name": "geomField",
                "type": "Point",  # optional
                "srs": "EPSG:4326",  # optional
            }
        ]  # Required, or geometry_fields() method

        self.metadata = {"foo": "bar"}  # optional

        # uncomment if __iter__() honour self.attribute_filter
        # self.iterator_honour_attribute_filter = True

        # uncomment if __iter__() honour self.spatial_filter
        # self.iterator_honour_spatial_filter = True

        # uncomment if feature_count() honour self.attribute_filter
        # self.feature_count_honour_attribute_filter = True

        # uncomment if feature_count() honour self.spatial_filter
        # self.feature_count_honour_spatial_filter = True

        # End of reserved attribute names

        self.count = 5

    # Required, unless self.name attribute is defined
    # def name(self):
    #    return 'my_layer'

    # Optional. If not defined, fid name is 'fid'
    # def fid_name(self):
    #    return 'my_fid'

    # Required, unless self.geometry_fields attribute is defined
    # def geometry_fields(self):
    #    return [...]

    # Required, unless self.required attribute is defined
    # def fields(self):
    #    return [...]

    # Optional. Only to be usd if self.metadata field is not defined
    # def metadata(self, domain):
    #    if domain is None:
    #        return {'foo': 'bar'}
    #    return None

    # Optional. Called when self.attribute_filter is changed by GDAL
    # def attribute_filter_changed(self):
    #     # You may change self.iterator_honour_attribute_filter
    #     # or feature_count_honour_attribute_filter
    #     pass

    # Optional. Called when self.spatial_filter is changed by GDAL
    # def spatial_filter_changed(self):
    #     # You may change self.iterator_honour_spatial_filter
    #     # or feature_count_honour_spatial_filter
    #     pass

    # Optional
    def test_capability(self, cap):
        if cap == BaseLayer.FastGetExtent:
            return True
        if cap == BaseLayer.StringsAsUTF8:
            return True
        # if cap == BaseLayer.FastSpatialFilter:
        #    return False
        # if cap == BaseLayer.RandomRead:
        #    return False
        if cap == BaseLayer.FastFeatureCount:
            return self.attribute_filter is None and self.spatial_filter is None
        return False

    # Optional
    def extent(self, force_computation):
        return [2.1, 49, 3, 50]  # minx, miny, maxx, maxy

    # Optional.
    def feature_count(self, force_computation):
        # As we did not declare feature_count_honour_attribute_filter and
        # feature_count_honour_spatial_filter, the below case cannot happen
        # But this is to illustrate that you can callback the default implementation
        # if needed
        # if self.attribute_filter is not None or \
        #   self.spatial_filter is not None:
        #    return super(Layer, self).feature_count(force_computation)

        return self.count

    # Required. You do not need to handle the case of simultaneous iterators on
    # the same Layer object.
    def __iter__(self):
        for i in range(self.count):
            properties = {
                "boolField": True,
                "int16Field": 32767,
                "int32Field": i + 2,
                "int64Field": 1234567890123,
                "realField": 1.23,
                "floatField": 1.2,
                "strField": "foo",
                "strNullField": None,
                "binaryField": b"\x01\x00\x02",
                "timeField": "12:34:56.789",
                "dateField": "2017-04-26",
                "datetimeField": "2017-04-26T12:34:56.789Z",
            }

            yield {
                "type": "OGRFeature",
                "id": i + 1,
                "fields": properties,
                "geometry_fields": {"geomField": "POINT(2 49)"},
                "style": "SYMBOL(a:0)" if i % 2 == 0 else None,
            }

    # Optional
    # def feature_by_id(self, fid):
    #    return {}


class Dataset(BaseDataset):

    # Optional, but implementations will generally need it
    def __init__(self, filename):
        # If the layers member is set, layer_count() and layer() will not be used
        self.layers = [Layer()]
        self.metadata = {"foo": "bar"}

    # Optional, called on native object destruction
    def close(self):
        pass

    # Optional. Only to be usd if self.metadata field is not defined
    # def metadata(self, domain):
    #    if domain is None:
    #        return {'foo': 'bar'}
    #    return None

    # Required, unless a layers attribute is set in __init__
    # def layer_count(self):
    #    return len(self.layers)

    # Required, unless a layers attribute is set in __init__
    # def layer(self, idx):
    #    return self.layers[idx]


# Required: class deriving from BaseDriver
class Driver(BaseDriver):

    # Optional. Called the first time the driver is loaded
    def __init__(self):
        pass

    # Required
    def identify(self, filename, first_bytes, open_flags, open_options={}):
        return filename == "DUMMY:"

    # Required
    def open(self, filename, first_bytes, open_flags, open_options={}):
        if not self.identify(filename, first_bytes, open_flags):
            return None
        return Dataset(filename)



#!/usr/bin/env python3

import sys

from osgeo.gdal import deprecation_warn

# import osgeo_utils.gdal_polygonize as a convenience to use as a script
from osgeo_utils.gdal_polygonize import *  # noqa
from osgeo_utils.gdal_polygonize import main

deprecation_warn("gdal_polygonize")
sys.exit(main(sys.argv))



#!/usr/bin/env python3

import glob
import os

options = {}


def collect_config_options(filename):
    lines = open(filename, "rt").readlines()
    i = -1
    len_lines = len(lines)

    func_names = ["CPLGetConfigOption"]
    if os.path.basename(filename) == "gt_overview.cpp":
        func_names.append("GetOptionValue")

    while True:
        i += 1
        if i >= len_lines:
            break
        l = lines[i][0:-1].strip()
        if l.startswith("/*"):
            continue

        found = False
        for func_name in func_names:
            pos = l.find(func_name + "(")
            if pos >= 0:
                pos_start = pos + len(func_name + "(")
                option = None
                if pos_start < len(l) and l[pos_start] == '"':
                    pos_end = l.find('"', pos_start + 1)
                    if pos_end + 1 < len(l) and l[pos_end + 1] == ",":
                        option = l[pos_start + 1 : pos_end]
                    else:
                        print("Skipping", l)
                elif pos_start == len(l):
                    l = lines[i + 1][0:-1].strip()
                    pos_start = l.find('"')
                    assert pos_start >= 0, l
                    pos_end = l.find('"', pos_start + 1)
                    if pos_end + 1 < len(l) and l[pos_end + 1] == ",":
                        option = l[pos_start + 1 : pos_end]
                    elif (
                        l
                        != """("OGR_" + GetDriverUCName() + "_USE_BBOX").c_str(), "YES")));"""
                    ):
                        print("Skipping", l)
                if option:
                    if func_name == "GetOptionValue":
                        option += "_OVERVIEW"
                    if option not in options:
                        options[option] = set([os.path.basename(filename)])
                    else:
                        options[option].add(os.path.basename(filename))

                found = True
                break

        if found:
            continue

        pos = l.find("VSIGetPathSpecificOption(")
        if pos >= 0 and l.find("VSIGetPathSpecificOption()") < 0:
            pos_start = pos + len("VSIGetPathSpecificOption(")
            option = None
            if pos_start == len(l):
                l += lines[i + 1][0:-1].strip()
                ++i

            pos_start2 = l.find(', "', pos_start)
            if pos_start2 < 0:
                l += " " + lines[i + 1][0:-1].strip()
                ++i

            pos_start2 = l.find(', "', pos_start)
            if pos_start2 < 0:
                if (
                    l.endswith(', std::string("VSI")')
                    or l.endswith("VSIGetPathSpecificOption(pszFilename, pszFilename,")
                    or l.find("VSIGetPathSpecificOption(pszPath, pszKey, pszDefault)")
                    >= 0
                    or l.find(
                        "VSIGetPathSpecificOption(osPathForOption.c_str(), osPathForOption.c_str(),"
                    )
                    >= 0
                    or l.find(
                        "VSIGetPathSpecificOption(pszFilename, sTuple.pszEnvVar, nullptr)"
                    )
                    >= 0
                    or l.find(
                        "VSIGetPathSpecificOption(osNonStreamingFilename.c_str(), sTuple.pszEnvVar, nullptr)"
                    )
                    >= 0
                    or l.find(
                        "const char *VSIGetPathSpecificOption(const char *pszPath, const char *pszKey, const char *pszDefault)"
                    )
                    >= 0
                ):
                    continue
            assert pos_start2 >= 0, l
            pos_end = l.find('"', pos_start2 + 3)
            if pos_end + 1 < len(l) and l[pos_end + 1] == ",":
                option = l[pos_start2 + 3 : pos_end]
            else:
                if (
                    l.find(
                        'VSIGetPathSpecificOption(osPathForOption.c_str(), pszOptionKey, "")'
                    )
                    >= 0
                    or l.find(
                        'VSIGetPathSpecificOption(osPathForOption.c_str(), pszURLKey, "")'
                    )
                    >= 0
                    or l.find(
                        'VSIGetPathSpecificOption(osPathForOption.c_str(), pszUserKey, "")'
                    )
                    >= 0
                    or l.find(
                        'VSIGetPathSpecificOption(osPathForOption.c_str(), pszPasswordKey, "")'
                    )
                    >= 0
                ):
                    continue
                print("Skipping", l)

            if option:
                if option not in options:
                    options[option] = set([os.path.basename(filename)])
                else:
                    options[option].add(os.path.basename(filename))

        pos = l.find("alt_config_option='")
        if pos >= 0:
            pos_start = pos + len("alt_config_option='")
            pos_end = l.find("'", pos_start)
            option = l[pos_start:pos_end]
            if option not in options:
                options[option] = set([os.path.basename(filename)])
            else:
                options[option].add(os.path.basename(filename))


def explore(dirname):
    for filename in glob.glob(dirname + "/*"):
        if os.path.isdir(filename):
            explore(filename)
        if filename.endswith(".c") or filename.endswith(".cpp"):
            collect_config_options(filename)


for dirname in ("alg", "port", "gcore", "frmts", "ogr", "apps", "swig"):
    explore(os.path.join(os.path.dirname(__file__), "..", dirname))

c_file = "// This file was automatically generated by collect_config_options.py\n"
c_file += "// DO NOT EDIT\n\n"
c_file += "// clang-format off\n"
c_file += "constexpr static const char* const apszKnownConfigOptions[] =\n"
c_file += "{\n"
for option in sorted(options, key=str.lower):
    filename = ", ".join(sorted(options[option]))
    c_file += f'   "{option}", // from {filename}\n'
c_file += "};\n"
c_file += "// clang-format on\n"
# print(c_file)

open(
    os.path.join(os.path.dirname(__file__), "..", "port", "cpl_known_config_options.h"),
    "wt",
).write(c_file)



# ******************************************************************************
#  Project:  GDAL ECW Driver
#  Purpose:  Script to lookup ECW (GDT) coordinate systems and translate
#            into OGC WKT for storage in $GDAL_HOME/data/ecw_cs.wkt.
#  Author:   Frank Warmerdam, warmerdam@pobox.com
#
# ******************************************************************************
#  Copyright (c) 2003, Frank Warmerdam
#
# SPDX-License-Identifier: MIT
# ******************************************************************************

import osr

##############################################################################
# rtod(): radians to degrees.


def r2d(rad):
    return float(rad) / 0.0174532925199433


##############################################################################
# load_dict()


def load_dict(filename):
    lines = open(filename).readlines()

    this_dict = {}
    for line in lines:
        if line[:8] != "proj_name":
            tokens = [token.strip() for token in line.split(",")]

            this_dict[tokens[0]] = tokens

    return this_dict


##############################################################################
# Mainline


# dir = 'M:/software/ER Viewer 2.0c/GDT_Data/'
directory = "/u/data/ecw/gdt/"

dict_list = [
    "tranmerc",
    "lambert1",
    "lamcon2",
    "utm",
    "albersea",
    "mercator",
    "obmerc_b",
    "grinten",
    "cassini",
    "lambazea",
    "datum_sp",
]

dict_dict = {}

for item in dict_list:
    dict_dict[item] = load_dict(directory + item + ".dat")
    # print 'loaded: %s' % item

pfile = open(directory + "project.dat")
pfile.readline()

for line in pfile.readlines():
    try:
        tokens = line.strip().split(",")
        if len(tokens) < 3:
            continue

        tokens = [token.strip() for token in tokens]

        ident = tokens[0]
        typ = tokens[1]

        lsize = float(tokens[2])
        lsize_str = tokens[2]

        dline = dict_dict[typ][ident]

        srs = osr.SpatialReference()

        # Handle translation of the projection parameters.

        if type != "utm":
            fn = float(dline[1])
            fe = float(dline[2])

        if type == "tranmerc":
            srs.SetTM(r2d(dline[5]), r2d(dline[4]), float(dline[3]), fe, fn)

        elif type == "mercator":
            srs.SetMercator(r2d(dline[5]), r2d(dline[4]), float(dline[3]), fe, fn)

        elif type == "grinten":
            srs.SetVDG(r2d(dline[3]), fe, fn)

        elif type == "cassini":
            srs.SetCS(r2d(dline[4]), r2d(dline[3]), fe, fn)

        elif type == "lambazea":
            srs.SetLAEA(r2d(dline[5]), r2d(dline[4]), fe, fn)

        elif type == "lambert1":
            srs.SetLCC1SP(r2d(dline[5]), r2d(dline[4]), float(dline[3]), fe, fn)

        elif type == "lamcon2":
            srs.SetLCC(
                r2d(dline[7]), r2d(dline[8]), r2d(dline[9]), r2d(dline[6]), fe, fn
            )

        # elif type == 'lambert2':
        #   false_en = '+y_0=%.2f +x_0=%.2f' \
        #   % (float(dline[12])*lsize, float(dline[13])*lsize)
        #     result = '+proj=lcc %s +lat_0=%s +lon_0=%s +lat_1=%s +lat_2=%s' \
        #           % (false_en, r2d(dline[3]), r2d(dline[4]),
        #           r2d(dline[7]), r2d(dline[8]))

        elif type == "albersea":
            srs.SetACEA(
                r2d(dline[3]), r2d(dline[4]), r2d(dline[5]), r2d(dline[6]), fe, fn
            )

        # elif type == 'obmerc_b':
        #     result = '+proj=omerc %s +lat_0=%s +lonc=%s +alpha=%s +k=%s' \
        #              % (false_en, r2d(dline[5]), r2d(dline[6]), r2d(dline[4]), dline[3])

        elif type == "utm":
            srs.SetUTM(int(dline[1]), dline[2] != "S")

        # Handle Units from projects.dat file.
        if srs.IsProjected():
            srs.SetAttrValue("PROJCS", ident)
            if lsize_str == "0.30480061":
                srs.SetLinearUnits("US Foot", float(lsize_str))
            elif lsize_str != "1.0":
                srs.SetLinearUnits("unnamed", float(lsize_str))

        wkt = srs.ExportToWkt()
        if wkt:
            print("%s,%s" % (ident, srs.ExportToWkt()))
        else:
            print('%s,LOCAL_CS["%s - (unsupported)"]' % (ident, ident))

    except KeyError:
        print('%s,LOCAL_CS["%s - (unsupported)"]' % (ident, ident))

    except Exception:
        print("can't translate: ", line)
        raise

# Translate datums to their underlying spheroid information.

pfile = open(directory + "datum.dat")
pfile.readline()

for line in pfile.readlines():
    tokens = [token.strip() for token in line.strip().split(",")]

    ident = tokens[0]

    sp_name = tokens[2]
    dline = dict_dict["datum_sp"][ident]
    srs = osr.SpatialReference()

    if ident == "WGS84":
        srs.SetWellKnownGeogCS("WGS84")
    elif ident == "NAD27":
        srs.SetWellKnownGeogCS("NAD27")
    elif ident == "NAD83":
        srs.SetWellKnownGeogCS("NAD83")
    else:
        srs.SetGeogCS(tokens[1], ident, sp_name, float(dline[2]), float(dline[4]))

    print("%s,%s" % (ident, srs.ExportToWkt()))
